@@ -1,18 +1,26 @@
-import pyfra.utils.misc
-import pyfra.shell
-from functools import wraps, partial
 
+import pyfra.remote
+import pyfra.shell
+
+__all__ = [
+    'apt',
+    'install_pyenv',
+    'ensure_supported',
+]
 
 def apt(r, packages):
-    # install sudo if it's not installed; this is the case in some docker containers
-    r.sh(f"sudo echo hi || {{ apt-get update; apt-get install sudo; }}; sudo apt-get update; sudo apt-get install -y {' '.join(packages)}", pyenv_version=None)
+    r.sh(f"sudo apt-get update; sudo apt-get install -y {' '.join(packages)}", pyenv_version=None)
 
 
 def ensure_supported(r):
+    # todo: wire this up to something
+    
     supported = [
         "Ubuntu 18", "Ubuntu 20", 
         "stretch"   # debian stretch
     ]
+
+    @pyfra.remote.block
     def _f(r):
         print("Checking if", r, "is running a supported distro")
 
@@ -21,7 +29,7 @@ def ensure_supported(r):
             for ver in supported
         ])
 
-    pyfra.utils.misc.once(partial(_f, r), name="ensure_supported;" + r.fingerprint)()
+    _f(r)
 
 ## things to install
 
