@@ -1,6 +1,6 @@
-import errno
+import json
+import pathlib
 import os
-import random
 import re
 import shlex
 import shutil
@@ -8,25 +8,29 @@ import subprocess
 import sys
 import time
 import urllib
-import json
 
 from best_download import download_file
 from colorama import Fore, Style
 from natsort import natsorted
+
+import imohash
 import pyfra.remote
-import re
+
+class ShellException(Exception):
+    def __init__(self, code, rem=False):
+        super().__init__(f"Command exited with non-zero error code {code}" + 
+            (". This could either be because the ssh connection could not be made, or because the command itself failed." if rem and code == 255 else ""))
+        self.returncode = code
 
 
-class ShellException(Exception): pass
-
-
-__all__ = ['sh', 'copy', 'ls', 'rm', 'curl', 'wget', 'quote']
+__all__ = ['sh', 'copy', 'ls', 'curl', 'quote', 'ShellException']
 
 
 def _wrap_command(x, no_venv=False, pyenv_version=None):
     bashrc_payload = r"""import sys,re; print(re.sub("If not running interactively.{,128}?esac", "", sys.stdin.read(), flags=re.DOTALL).replace('[ -z "$PS1" ] && return', ''))"""
     hdr = f"shopt -s expand_aliases; ctrlc() {{ echo Shell wrapper interrupted with C-c, raising error; exit 174; }}; trap ctrlc SIGINT; "
-    hdr += f"eval \"$(cat ~/.bashrc | python3 -c {bashrc_payload | quote})\"  > /dev/null 2>&1; "
+    hdr += f"[ -e ~/.bashrc ] && {{ eval \"$(cat ~/.bashrc | python3 -c {bashrc_payload | quote})\"  > /dev/null 2>&1; }}; "
+    hdr += "[ -e ~/.bashrc ] || { [ -e ~/.zshrc ] && . ~/.zshrc; }; "
     hdr += "python() { python3 \"$@\"; };" # use python3 by default
     if pyenv_version is not None: hdr += f"pyenv shell {pyenv_version} || exit 1 > /dev/null 2>&1; "
     if not no_venv: hdr += "[ -f env/bin/activate ] && . env/bin/activate; "
@@ -121,10 +125,13 @@ def sh(cmd, quiet=False, wd=None, wrap=True, maxbuflen=1000000000, ignore_errors
     """
     if wd is None: wd = os.getcwd()
 
-    return _rsh("127.0.0.1", cmd, quiet, wd, wrap, maxbuflen, -1, ignore_errors, no_venv, pyenv_version)
+    try:
+        return _rsh("127.0.0.1", cmd, quiet, wd, wrap, maxbuflen, -1, ignore_errors, no_venv, pyenv_version)
+    except ShellException as e: # this makes the stacktrace easier to read
+        raise ShellException(e.returncode) from None
 
 
-def _rsh(host, cmd, quiet=False, wd=None, wrap=True, maxbuflen=1000000000, connection_timeout=10, ignore_errors=False, no_venv=False, pyenv_version=None):
+def _rsh(host, cmd, quiet=False, wd=None, wrap=True, maxbuflen=1000000000, connection_timeout=10, ignore_errors=False, no_venv=False, pyenv_version=None, forward_keys=False):
     if host is None or host == "localhost": host = "127.0.0.1"
 
     # implicit-copy files from remote to local
@@ -152,8 +159,9 @@ def _rsh(host, cmd, quiet=False, wd=None, wrap=True, maxbuflen=1000000000, conne
 
     if wrap: cmd = _wrap_command(cmd, no_venv=no_venv, pyenv_version=pyenv_version)
     if wd: cmd = f"cd {wd}  > /dev/null 2>&1; {cmd}"
-
-    ret = _sh(f"ssh -q -oConnectTimeout={connection_timeout} -oBatchMode=yes -oStrictHostKeyChecking=no -oUserKnownHostsFile=/dev/null -t {host} {shlex.quote(cmd)}", quiet=quiet, wrap=False, maxbuflen=maxbuflen, ignore_errors=ignore_errors, no_venv=no_venv)
+ 
+    ssh_cmd = "eval \"$(ssh-agent -s)\"; ssh-add ~/.ssh/id_rsa; ssh -A" if forward_keys else "ssh"
+    ret = _sh(f"{ssh_cmd} -q -oConnectTimeout={connection_timeout} -oBatchMode=yes -oStrictHostKeyChecking=no -oUserKnownHostsFile=/dev/null -t {host} {shlex.quote(cmd)}", quiet=quiet, wrap=False, maxbuflen=maxbuflen, ignore_errors=ignore_errors, no_venv=no_venv)
 
     # implicit-copy files from local to remote
     for remf, locf, copyerr in rempaths:
@@ -166,7 +174,7 @@ def _rsh(host, cmd, quiet=False, wd=None, wrap=True, maxbuflen=1000000000, conne
 
     return ret
 
-def copy(frm, to, quiet=False, connection_timeout=10, symlink_ok=True, into=True, exclude=[]):
+def copy(frm, to, quiet=False, connection_timeout=10, symlink_ok=True, into=True, exclude=[]) -> None:
     """
     Copies things from one place to another.
 
@@ -178,42 +186,72 @@ def copy(frm, to, quiet=False, connection_timeout=10, symlink_ok=True, into=True
         symlink_ok (bool): If frm and to are on the same machine, symlinks will be created instead of actually copying. Set to false to force copying.
         into (bool): If frm is a file, this has no effect. If frm is a directory, then into=True for frm="src" and to="dst" means "src/a" will get copied to "dst/src/a", whereas into=False means "src/a" will get copied to "dst/a".
     """
-    if isinstance(frm, pyfra.remote.RemotePath): frm = frm.rsyncstr()
-    if isinstance(to, pyfra.remote.RemotePath): to = to.rsyncstr()
-
-    if not quiet: print(f"{Style.BRIGHT}{Fore.RED}*{Style.RESET_ALL} Copying {Style.BRIGHT}{frm} {Style.RESET_ALL}to {Style.BRIGHT}{to}{Style.RESET_ALL}")
 
     # copy from url
-    if frm.startswith("http://") or frm.startswith("https://"):
+    if isinstance(frm, str) and (frm.startswith("http://") or frm.startswith("https://")):
         if ":" in to:
             to_host, to_path = to.split(":")
-            _rsh(to_host, f"curl {frm} --create-dirs -o {to}")
+            _rsh(to_host, f"curl {frm} --create-dirs -o {to_path}")
         else:
             wget(frm, to)
         return
 
-    if frm[-1] == '/' and len(frm) > 1: frm = frm[:-1]
-    if not into: frm += '/'
+    # get rsync strs and make sure frm and to are RemotePaths
+    if isinstance(frm, pyfra.remote.RemotePath): 
+        frm_str = frm.rsyncstr()
+    else:
+        frm_str = frm
+        assert ":" not in frm_str
+        frm = pyfra.remote.local.path(frm)
+
+    if isinstance(to, pyfra.remote.RemotePath):
+        to_str = to.rsyncstr()
+    else:
+        to_str = to
+        assert ":" not in to_str
+        to = pyfra.remote.local.path(to)
+    
+    # state tracking
+    needs_set_kv = False
+    if not to.remote._no_hash:
+        with to.remote.no_hash():
+            checksum = frm.quick_hash()
+            new_hash = to.remote.update_hash("copy", to.fname, checksum)
+            try:
+                to.remote.get_kv(new_hash)
+                # if already copied, then return
+                to._set_cache("quick_hash", checksum) # set the checksum of the target file to avoid needing to calculate it again
+                pyfra.remote._print_skip_msg(to.remote.envname, "copy", new_hash)
+
+                return
+            except KeyError:
+                needs_set_kv = True
+
+    # print info
+    if not quiet: print(f"{Style.BRIGHT}{Fore.RED}*{Style.RESET_ALL} Copying {Style.BRIGHT}{frm_str} {Style.RESET_ALL}to {Style.BRIGHT}{to_str}{Style.RESET_ALL}")
+
+    if frm_str[-1] == '/' and len(frm_str) > 1: frm_str = frm_str[:-1]
+    if not into: frm_str += '/'
 
     if quiet:
         opts = "-e \"ssh -o StrictHostKeyChecking=no\" -arqL"
     else:
-        opts = "-e \"ssh -o StrictHostKeyChecking=no\" -arL --info=progress2"
+        opts = "-e \"ssh -o StrictHostKeyChecking=no\" -arL"
     
     for ex in exclude:
         opts += f" --exclude {ex | pyfra.shell.quote}"
     
-    def symlink_frm(frm):
-        # rsync behavior is to copy the contents of frm into to if frm ends with a /
-        if frm[-1] == '/': frm += '*'
+    def symlink_frm(frm_str):
+        # rsync behavior is to_str copy the contents of frm_str into to_str if frm_str ends with a /
+        if frm_str[-1] == '/': frm_str += '*'
         # ln -s can't handle relative paths well! make absolute if not already
-        if frm[0] != '/' and frm[0] != '~': frm = "$PWD/" + frm
+        if frm_str[0] != '/' and frm_str[0] != '~': frm_str = "$PWD/" + frm_str
 
-        return frm
+        return frm_str
 
-    if ":" in frm and ":" in to:
-        frm_host, frm_path = frm.split(":")
-        to_host, to_path = to.split(":")
+    if ":" in frm_str and ":" in to_str:
+        frm_host, frm_path = frm_str.split(":")
+        to_host, to_path = to_str.split(":")
 
         par_target = to_path.rsplit('/', 1)[0] if "/" in to_path else ""
 
@@ -226,23 +264,29 @@ def copy(frm, to, quiet=False, connection_timeout=10, symlink_ok=True, into=True
                 if par_target: _rsh(to_host, f"mkdir -p {par_target}", quiet=True)
                 _rsh(frm_host, f"rsync {opts} {frm_path} {to_path}", quiet=True)
         else:
-            rsync_cmd = f"rsync {opts} {frm_path} {to}"
+            rsync_cmd = f"rsync {opts} {frm_path} {to_str}"
                 
             # make parent dir in terget if not exists
             if par_target: _rsh(to_host, f"mkdir -p {par_target}", quiet=True)
 
             sh(f"eval \"$(ssh-agent -s)\"; ssh-add ~/.ssh/id_rsa; ssh -q -oConnectTimeout={connection_timeout} -oBatchMode=yes -oStrictHostKeyChecking=no -oUserKnownHostsFile=/dev/null -A {frm_host} {rsync_cmd | quote}", wrap=False, quiet=True)
     else:
-        # if to is host:path, then this gives us path; otherwise, it leaves it unchanged
-        par_target = to.split(":")[-1]
+        # if to_str is host:path, then this gives us path; otherwise, it leaves it unchanged
+        par_target = to_str.split(":")[-1]
         par_target = par_target.rsplit('/', 1)[0] if "/" in par_target else ""
 
-        if symlink_ok and ":" not in frm and ":" not in to:
+        if symlink_ok and ":" not in frm_str and ":" not in to_str:
             assert not exclude, "Cannot use exclude symlink"
-            sh(f"[ -d {frm} ] && mkdir -p {par_target}; ln -sf {symlink_frm(frm)} {to}", quiet=True)
+            sh(f"[ -d {frm_str} ] && mkdir -p {par_target}; ln -sf {symlink_frm(frm_str)} {to_str}", quiet=True)
         else:
-            if ":" in to: _rsh(to.split(":")[0], f"mkdir -p {par_target}", quiet=True)
-            sh((f"mkdir -p {par_target}; " if par_target and ":" in frm else "") + f"rsync {opts} {frm} {to}", wrap=False, quiet=False)
+            if ":" in to_str: _rsh(to_str.split(":")[0], f"mkdir -p {par_target}", quiet=True)
+            sh((f"mkdir -p {par_target}; " if par_target and ":" in frm_str else "") + f"rsync {opts} {frm_str} {to_str}", wrap=False, quiet=True)
+    
+    # set value in key value store to flag as done
+    if needs_set_kv:
+        to.remote.set_kv(new_hash, None)
+        to._set_cache("quick_hash", checksum) # set the checksum of the target file to avoid needing to calculate it again
+
 
 def ls(x='.'):
     return list(natsorted([x + '/' + fn for fn in os.listdir(x)]))
@@ -285,6 +329,19 @@ def wget(url, to=None, checksum=None):
 
     download_file(url, to, checksum)
 
+def quick_hash(path):
+    params = {
+        "hexdigest": True,
+        "sample_size": 4 * 1024**2, # 4 MB
+        "sample_threshhold": 16 * 1024**2, # 16 MB
+    }
+    path = os.path.expanduser(path)
+    if pathlib.Path(path).is_dir():
+        files = list(sorted(pathlib.Path(path).glob('**/*')))
+        res = pyfra.remote._hash_obs(*[(str(f.relative_to(pathlib.Path(path))), imohash.hashfile(str(f.resolve()), **params)) for f in files if f.is_file()])[:32]
+        return res
+    return imohash.hashfile(path, **params)
+
 # convenience function for shlex.quote
 class _quote:
     def __ror__(self, other):
